/*
 *        spOption.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>
#include <ctype.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spFile.h>
#include <sp/spOption.h>

#if defined(MACOSX)
#if defined(IPHONE)
#include <CoreFoundation/CoreFoundation.h>
#else
#include <CoreServices/CoreServices.h>
#endif
#endif

extern int spvsnprintf(char *buf, size_t n, const char *format, va_list args);

struct _spOptions {
    char *progname;        /* program name */
    int section;        /* section number */
    int num_option;        /* number of option */
    spOption *options;        /* option list */
    spBool option_allocated;
    int num_file;        /* number of file */
    const char **labels;        /* file label */
    char **files;        /* file name */
    spBool *changes;
    int argc;
    int filec;
    int file_index;
    spBool ignore_unknown;
};

static spOptions sp_options = NULL;
static char *sp_setup_file = NULL;
static spBool *sp_help_flag;
static char sp_help_message[SP_MAX_LINE] = "";
static int sp_option_desc_offset = 32;

spBool spEqType(spOptionType type1, spOptionType type2)
{
    enum {SP_DUMMY1, SP_DUMMY2} dummy;
    
    if (!(0xff & type2)) {
        if ((0xff00 & type2) & (0xff00 & type1)) {
            return SP_TRUE;
        }
    } else {
        if ((0xff & type1) == SP_TYPE_ENUM) {
            if (sizeof(dummy) == sizeof(char)) {
                return ((0xff & type2) == SP_TYPE_CHAR ? SP_TRUE : SP_FALSE);
            } else if (sizeof(dummy) == sizeof(short)) {
                return ((0xff & type2) == SP_TYPE_SHORT ? SP_TRUE : SP_FALSE);
            } else if (sizeof(dummy) == sizeof(long)) {
                return ((0xff & type2) == SP_TYPE_LONG ? SP_TRUE : SP_FALSE);
            } else {
                return ((0xff & type2) == SP_TYPE_INT ? SP_TRUE : SP_FALSE);
            }
        } else if ((0xff & type1) == (0xff & type2)) {
            return SP_TRUE;
        }
    }

    return SP_FALSE;
}

void spIgnoreUnknownOption(spOptions options, spBool flag)
{
    if (options == NULL) return;
    
    options->ignore_unknown = flag;
    return;
}

int spGetNumOption(spOptions options)
{
    if (options == NULL) return -1;
    
    return options->num_option;
}

spOption *spGetOptionList(spOptions options)
{
    if (options == NULL) return NULL;
    
    return options->options;
}

/*
 *        cut option value
 */
char *xspCutOptionValue(const char *value, int index)
{
    int i;
    char *p;
    char *string = NULL;
    char buf[SP_MAX_LINE];

    if (strnone(value)) return NULL;

    p = (char *)value;
    for (i = 0;; i++) {
        if (index < 0) {
            if (*p == '@') {
                string = p + 1;
                break;
            }
        } else {
            if (i == index) {
                if (*p == '@') {
                    string = p + 1;
                } else {
                    string = p;
                }
                break;
            }
        }
        if ((p = strchr(p, '|')) == NULL) {
            break;
        }
        p++;
    }

    if (index < 0 && string == NULL) {
        string = (char *)value;
    }
    
    if (string != NULL) {
        spStrCopy(buf, sizeof(buf), string);
    
        if ((p = strchr(buf, '|')) != NULL) {
            *p = NUL;
        }

        return strclone(buf);
    } else {
        return NULL;
    }
}

/*
 *        convert option value
 */
static int convertOptionValue(spOption *option, const char *value, int init_flag)
{
    int incr = SP_UNKNOWN;
    char *string;

    if (option == NULL) return SP_UNKNOWN;

    spDebug(40, "convertOptionValue", "in\n");
    
    if (value == NULL || *value == NUL) {
        spDebug(40, "convertOptionValue", "value is NULL\n");
        if (option->value != NULL) {
            if (spEqType(option->type, SP_TYPE_STRING)) {
                if (!init_flag) {
                    if (*(char **)option->value != NULL)
                        xfree(*(char **)option->value);
                }
                *(char **)option->value = NULL;
            } else if (spEqType(option->type,  SP_TYPE_STRING_A)) {
                spStrCopy((char *)option->value, 1, "");
            }
        }
    } else {
        spDebug(40, "convertOptionValue", "value = %s\n", value);
        
        if ((string = xspCutOptionValue(value, -1)) != NULL) {
            spDebug(60, "convertOptionValue", "string = %s\n", string);
            if (option->value == NULL) {
                if (spEqType(option->type, SP_TYPE_BOOL)) {
                    incr = 0;
                } else {
                    incr = 1;
                }
            } else if (string != NULL) {
                if (spEqType(option->type, SP_TYPE_INT)) {
                    *(int *)option->value = atoi(string);
                    incr = 1;
                } else if (spEqType(option->type, SP_TYPE_SHORT)) {
                    *(short *)option->value = (short)atoi(string);
                    incr = 1;
                } else if (spEqType(option->type, SP_TYPE_LONG)) {
                    *(long *)option->value = atol(string);
                    spDebug(40, "convertOptionValue", "SP_TYPE_LONG: string = %s, value = %ld\n", string, *(long *)option->value);
                    incr = 1;
                } else if (spEqType(option->type, SP_TYPE_FLOAT)) {
                    *(float *)option->value = (float)atof(string);
                    incr = 1;
                } else if (spEqType(option->type, SP_TYPE_DOUBLE)) {
                    *(double *)option->value = atof(string);
                    incr = 1;
                } else if (spEqType(option->type, SP_TYPE_CHAR)) {
                    *(char *)option->value = string[0];
                    incr = 1;
                } else if (spEqType(option->type, SP_TYPE_STRING)) {
                    if (!init_flag) {
                        if (*(char **)option->value != NULL) {
                            xfree(*(char **)option->value);
                        }
                    }
                    *(char **)option->value = strclone(string);
                    spDebug(40, "convertOptionValue", "SP_TYPE_STRING: init_flag = %d, value = %s\n",
                            init_flag, *(char **)option->value);
                    incr = 1;
                } else if (spEqType(option->type, SP_TYPE_STRING_A)) {
                    spStrCopy((char *)option->value, (int)strlen(string) + 1, string);
                    incr = 1;
                } else if (spEqType(option->type, SP_TYPE_BOOL)) {
                    *(spBool *)option->value = spStrToBool(string);
                    incr = 0;
                } else {
                    spWarning("Unknown option data type.\n");
                    incr = SP_UNKNOWN;
                }
            } else {
                incr = SP_UNKNOWN;
            }
            xfree(string);
        } else {
            spDebug(40, "convertOptionValue", "xspCutOptionValue failed\n");
        }
    }

    spDebug(40, "convertOptionValue", "done\n");
    
    return incr;
}

int spConvertOptionValue(spOption *option, const char *value)
{
    return convertOptionValue(option, value, 0);
}

/*
 *        initialize option
 */
static void initOptions(spOptions options)
{
    int j;

    for (j = 0; j < options->num_option; j++) {
        convertOptionValue(&options->options[j], options->options[j].def_value, 1);
    }

    spDebug(40, "initOptions", "done\n");
    
    return;
}

/*
 *      get option number of being equal to flag
 */
static int flagEq(spOptions options, char *flag)
{
    int i;

    spDebug(40, "flagEq", "flag = %s\n", flag);
    
    for (i = 0;; i++) {
        if (i >= options->num_option) {
            return SP_UNKNOWN;
        }
        if (streq(options->options[i].flag, flag) ||
            streq(options->options[i].subflag, flag)) {
            break;
        } else if (options->options[i].type == SP_TYPE_BOOL && flag[0] == '+') {
            if ((options->options[i].flag != NULL 
                 && streq(options->options[i].flag + 1, flag + 1)) ||
                (options->options[i].subflag != NULL 
                 && streq(options->options[i].subflag + 1, flag + 1))) {
                break;
            }
        }
    }

    return i;
}

/*
 *        set change flag
 */
static void setChanges(int argc, char **argv, spOptions options)
{
    int i;
    int oc;

    if (options->changes == NULL) {
        options->changes = xalloc(MAX(options->num_option, 1), spBool);
        for (oc = 0; oc < options->num_option; oc++) {
            options->changes[oc] = SP_FALSE;
        }
    }
    
    for (i = 1; i < argc; i++) {
        if ((oc = flagEq(options, argv[i])) != SP_UNKNOWN) {
            options->changes[oc] = SP_TRUE;
        }
    }

    return;
}

static spOptions allocOptions(int num_option, spOption *option_list, spBool copy_flag)
{
    spOptions options;
    
    options = xalloc(1, struct _spOptions);
    options->progname = NULL;
    options->section = 1;
    options->num_option = num_option;
    if (copy_flag == SP_TRUE) {
        int i;
        
        options->options = xalloc(num_option, spOption);
        for (i = 0; i < num_option; i++) {
            memcpy(&options->options[i], &option_list[i], sizeof(spOption));
        }
        options->option_allocated = SP_TRUE;
    } else {
        options->options = option_list;
        options->option_allocated = SP_FALSE;
    }
    options->num_file = 0;
    options->labels = NULL;
    options->files = NULL;
    options->changes = NULL;
    options->argc = 1;
    options->filec = 0;
    options->file_index = 0;
    options->ignore_unknown = SP_FALSE;

    return options;
}

spOptions spAllocOptions(int num_option, spOption *option_list)
{
    return allocOptions(num_option, option_list, SP_FALSE);
}

spOptions spCopyOptions(int num_option, spOption *option_list)
{
    return allocOptions(num_option, option_list, SP_TRUE);
}

spOptions spInitOptions(int num_option, spOption *option_list)
{
    spOptions options;
    
    if ((options = allocOptions(num_option, option_list, SP_TRUE)) != NULL) {
        initOptions(options);
        return options;
    }

    return NULL;
}

void _spFreeOptionsAndValues(spOptions options, spBool free_values)
{
    int i;
    
    if (options != NULL) {
        if (options == sp_options) {
            sp_options = NULL;
        }

        if (free_values) {
            for (i = 0; i < options->num_option; i++) {
                if (spEqType(options->options[i].type, SP_TYPE_STRING)) {
                    if (options->options[i].value != NULL && *(char **)options->options[i].value != NULL) {
                        xfree(*(char **)options->options[i].value);
                        *(char **)options->options[i].value = NULL;
                    }
                }
            }
        }
        
        if (options->option_allocated == SP_TRUE) {
            xfree(options->options);
        }
        if (options->progname != NULL) {
            xfree(options->progname);
        }
        if (options->files != NULL) {
            xfree(options->files);
        }
        if (options->changes != NULL) {
            xfree(options->changes);
        }
        xfree(options);
    }

    return;
}

void _spFreeOptions(spOptions options)
{
    _spFreeOptionsAndValues(options, SP_FALSE);
}

void _spFreeOptionsOnExit(spOptions options)
{
    _spFreeOptionsAndValues(options, SP_TRUE);
}

/*
 *        get options struct
 */
spOptions _spGetOptions(int argc, char **argv, int num_option, spOption *option_list,
                        int num_file, const char **labels)
{
    int i;
    spOptions options;

    options = spAllocOptions(num_option, option_list);
    options->progname = xspGetBaseName(argv[0]);
    if (num_file == 1 && labels != NULL && strnone(labels[0])) {
        options->num_file = 0;
    } else {
        options->num_file = num_file;
    }
    options->labels = labels;
    if (argc > 0) {
        options->files = xalloc(argc, char *);
        for (i = 0; i < argc; i++) {
            options->files[i] = NULL;
        }
    }

    initOptions(options);
    setChanges(argc, argv, options);
    
    if (sp_options == NULL) {
        sp_options = options;
        spAddExitCallback((spExitCallbackFunc)_spFreeOptionsOnExit, sp_options);
    }

    spDebug(50, "_spGetOptions", "sp_setup_file = %s\n", sp_setup_file);
    
    if (!strnone(sp_setup_file)) {
        if (spReadSetup(sp_setup_file, options) == SP_TRUE) {
            spDebug(50, "_spGetOptions", "spReadSetup OK: %s\n", sp_setup_file);
        } else {
            char *basename;

            if ((basename = spGetBaseName(sp_setup_file)) != NULL) {
                char buf[SP_MAX_PATHNAME];
                
#if defined(HAVE_SPRINTF_S)
                sprintf_s(buf, sizeof(buf), "%s%c%s", spGetDefaultDir(), SP_DIR_SEPARATOR, basename);
#elif defined(HAVE_SNPRINTF)
                snprintf(buf, sizeof(buf), "%s%c%s", spGetDefaultDir(), SP_DIR_SEPARATOR, basename);
#else
                sprintf(buf, "%s%c%s", spGetDefaultDir(), SP_DIR_SEPARATOR, basename);
#endif
                if (spExists(buf) == SP_TRUE) {
                    spReadSetup(buf, options);
                }
            }
        }
    }

    return options;
}

/*
 *        set section number 
 */
void spSetSection(spOptions options, int section)
{
    if (options != NULL) {
        options->section = section;
    }
    
    return;
}

void spSetSetup(const char *filename)
{
    if (sp_setup_file != NULL) {
        xfree(sp_setup_file);
        sp_setup_file = NULL;
    }

    if (!strnone(filename)) {
        spDebug(20, "spSetSetup", "filename = %s\n", filename);
#if 0
        sp_setup_file = strclone(filename);
#else
        {
            char *basename;

            if (spIsExactName(filename) == SP_FALSE
                && (basename = spGetBaseName(filename)) != NULL) {
                int size;
                const char *appdir;

                appdir = spGetApplicationDir(NULL);

                size = (int)(strlen(appdir) + strlen(basename)) + 2;
                spDebug(100, "spSetSetup", "size = %d, appdir = %s, basename = %s\n", size, appdir, basename);
                sp_setup_file = xalloc(size, char);
#if defined(HAVE_SPRINTF_S)
                sprintf_s(sp_setup_file, size, "%s%c%s", appdir, SP_DIR_SEPARATOR, basename);
#elif defined(HAVE_SNPRINTF)
                snprintf(sp_setup_file, size, "%s%c%s", appdir, SP_DIR_SEPARATOR, basename);
#else
                sprintf(sp_setup_file, "%s%c%s", appdir, SP_DIR_SEPARATOR, basename);
#endif
            } else {
                sp_setup_file = strclone(filename);
            }
        }
#endif
        spDebug(80, "spSetSetup", "sp_setup_file = %s\n", sp_setup_file);
    }
    
    return;
}

spBool spWriteGlobalSetup(void)
{
    return spWriteSetup(sp_setup_file, sp_options);
}

typedef struct _spExitCallbackList *spExitCallbackList;
struct _spExitCallbackList {
    int num_callback_buffer;
    int num_callback;
    spExitCallbackFunc *callbacks;
    void  **call_data_list;
};

static spExitCallbackList sp_exit_callback_list = NULL;

spBool spAddExitCallback(spExitCallbackFunc callback, void *data)
{
    int index;
    
    if (callback == NULL) return SP_FALSE;
    
    if (sp_exit_callback_list == NULL) {
        sp_exit_callback_list = xalloc(1, struct _spExitCallbackList);
        sp_exit_callback_list->num_callback_buffer = 0;
        sp_exit_callback_list->num_callback = 0;
        sp_exit_callback_list->callbacks = NULL;
        sp_exit_callback_list->call_data_list = NULL;
    }

    if (sp_exit_callback_list == NULL) return SP_FALSE;

#if 0
    for (index = 0; index < sp_exit_callback_list->num_callback; index++) {
        if (sp_exit_callback_list->callbacks[index] == NULL) {
            break;
        }
    }
#else
    index = sp_exit_callback_list->num_callback;
#endif
    
    spDebug(20, "spAddExitCallback", "index = %d, num_callback = %d\n",
            index, sp_exit_callback_list->num_callback);

    if (index >= sp_exit_callback_list->num_callback) {
        if (sp_exit_callback_list->num_callback + 1 >= sp_exit_callback_list->num_callback_buffer) {
            sp_exit_callback_list->num_callback_buffer += SP_MAX_EXIT_CALLBACK;
            sp_exit_callback_list->callbacks = xrealloc(sp_exit_callback_list->callbacks,
                                                        sp_exit_callback_list->num_callback_buffer, spExitCallbackFunc);
            sp_exit_callback_list->call_data_list = xrealloc(sp_exit_callback_list->call_data_list,
                                                             sp_exit_callback_list->num_callback_buffer, void *);
        }
        index = sp_exit_callback_list->num_callback;
        sp_exit_callback_list->num_callback++;
    }

    sp_exit_callback_list->callbacks[index] = callback;
    sp_exit_callback_list->call_data_list[index] = data;

    spDebug(20, "spAddExitCallback", "done: index = %d, num_callback = %d\n",
            index, sp_exit_callback_list->num_callback);
    
    return SP_TRUE;
}

spBool spRemoveExitCallback(spExitCallbackFunc callback, void *data)
{
    int i;
    
    if (callback == NULL || sp_exit_callback_list == NULL) return SP_FALSE;
    
    for (i = 0; i < sp_exit_callback_list->num_callback; i++) {
        if (sp_exit_callback_list->callbacks[i] == callback
            && sp_exit_callback_list->call_data_list[i] == data) {
            sp_exit_callback_list->callbacks[i] = NULL;
            sp_exit_callback_list->call_data_list[i] = NULL;
            return SP_TRUE;
        }
    }
    
    return SP_FALSE;
}

static spExitFunc sp_exit_func = NULL;

spBool spSetExitFunc(spExitFunc func)
{
    sp_exit_func = func;
    return SP_TRUE;
}

spBool spEmitExitCallback(void)
{
    int i;
    spExitCallbackFunc callback;

    if (sp_exit_callback_list != NULL) {
        for (i = sp_exit_callback_list->num_callback - 1; i >= 0; i--) {
            if ((callback = sp_exit_callback_list->callbacks[i]) != NULL) {
                callback(sp_exit_callback_list->call_data_list[i]);
            }
        }
        if (sp_exit_callback_list->num_callback_buffer > 0) {
            xfree(sp_exit_callback_list->callbacks);
            xfree(sp_exit_callback_list->call_data_list);
        }
        xfree(sp_exit_callback_list);
        sp_exit_callback_list = NULL;
    }
    
    return SP_TRUE;
}

/*
 *        exit program
 */
void spExit(int status)
{
    if (status == 0) {
        spWriteGlobalSetup();
    }
    if (sp_setup_file != NULL) {
        xfree(sp_setup_file);
        sp_setup_file = NULL;
    }

    spEmitExitCallback();

    if (sp_exit_func != NULL) {
        sp_exit_func(status);
    } else {
        exit(status);
    }
}

/*
 *        set option value
 */
static int setOptionValue(spOption *option, char *flag, char *value)
{
    int incr;

    if (spEqType(option->type, SP_TYPE_BOOL)) {
        if (option->value != NULL) {
            if (*(spBool *)option->value != SP_TRUE) {
                if (flag != NULL && flag[0] == '+') {
                    *(spBool *)option->value = SP_FALSE;
                } else {
                    *(spBool *)option->value = SP_TRUE;
                }
            } else {
                if (flag != NULL && flag[0] == '+') {
                    *(spBool *)option->value = SP_TRUE;
                } else {
                    *(spBool *)option->value = SP_FALSE;
                }
            }
        }
        incr = 0;
    } else {
        incr = spConvertOptionValue(option, value);
    }

    spDebug(40, "setOptionValue", "done\n");
    
    return incr;
}

/*
 *        get arg file
 */
static int getArgFile(spOptions options, char *filename, int *fc)
{
    int i;
    int stdin_flag = 0;
    int incr = 1;

    spDebug(40, "getArgFile", "in\n");
    
    if (streq(filename, "-")) {
        stdin_flag = 1;
    }
    
    if (!stdin_flag && strveq(filename, "-")) {
        if (options->ignore_unknown == SP_TRUE) {
            return 0;
#if defined(MACOSX)
        } else if (strveq(filename, "-psn")
                   || strveq(filename, "-NS")) {
            return 0;
#endif
        } else {
            spPrintError("Unknown option %s", filename);
        }
    }
    if (fc == NULL) {
        i = 0;
    } else {
        i = *fc;
        *fc += incr;
    }

    if (options->files != NULL) {
        if (options->files[i] != NULL)
            xfree(options->files[i]);
        
        if (stdin_flag) {
            options->files[i] = strclone(filename);
        } else {
            options->files[i] = xspGetExactName(filename);
        }
        
        spDebug(20, "getArgFile", "options->files[%d] = %s\n", i,
                options->files[i]);
    }
    
    return incr;
}

/*
 *        get option value
 */
char *spGetOptionValue(int argc, char **argv, spOptions options)
{
    int i;
    int oc;
    int incr;

    if (argc <= 0 || argv == NULL || options == NULL) return NULL;

    i = options->argc;
    
    spDebug(40, "spGetOptionValue", "in: argc = %d, i = %d\n", argc, i);
    
    if (i >= argc)
        return NULL;

    if ((oc = flagEq(options, argv[i])) != SP_UNKNOWN) {
        if (i + 1 >= argc) {
            incr = setOptionValue(&(options->options[oc]), argv[i], (char *)NULL);
        } else {
            incr = setOptionValue(&(options->options[oc]), argv[i], argv[i + 1]);
        }
        if (incr == SP_UNKNOWN)
            spUsage();
    } else {
        getArgFile(options, argv[i], &options->filec);
        incr = 0;
    }

    options->argc += incr + 1;
    
    if (!strnone(sp_help_message) && *sp_help_flag == SP_TRUE)
        spPrintHelp(sp_help_message);
    
    return argv[i];
}

void spGetOptionsValue(int argc, char **argv, spOptions options)
{
    if (options == NULL) return;
    
    while (spGetOptionValue(argc, argv, options) != NULL);
    
    spDebug(40, "spGetOptionsValue", "spGetOptionValue done\n");

#if 0
    if (options->filec < options->num_file - 1) {
        spPrintError("Not enough files.");
    }
#endif
    
    return;
}

void spUpdateOptionsValue(int argc, char **argv, spOptions options)
{
    int i;
    int oc;
    int incr;
    
    if (argc <= 0 || argv == NULL || options == NULL) {
        return;
    }

    i = 0;
    while (1) {
        if (i >= argc) {
            break;
        }
        
        if ((oc = flagEq(options, argv[i])) != SP_UNKNOWN) {
            if (i + 1 >= argc) {
                incr = setOptionValue(&(options->options[oc]), argv[i], (char *)NULL);
            } else {
                incr = setOptionValue(&(options->options[oc]), argv[i], argv[i + 1]);
            }
            if (incr == SP_UNKNOWN) {
                incr = 0;
            }
        } else {
            incr = 0;
        }
        i += incr + 1;
    }

    return;
}

/*
 *        check number of file
 */
void spCheckNumFile(spOptions options)
{
    if (options == NULL) return;
    
    if (options->filec < options->num_file) {
        spPrintError("Not enough files.");
    } else if (options->filec > options->num_file) {
        spPrintError("Too many files.");
    }
}

/*
 *        get filename from command line
 */
const char *spGetFile(spOptions options)
{
    char *file = NULL;

    if (options == NULL) return NULL;
    
    spDebug(20, "spGetFile", "options->file_index = %d, options->num_file = %d\n",
            options->file_index, options->num_file);
    
    if (options->files != NULL
        && options->files[options->file_index] != NULL) {
        file = options->files[options->file_index];
        options->file_index++;
    }

    return file;
}

char *xspGetOptionLabel(spOption *option, int index)
{
    int i;
    char *p;
    char *label = NULL;
    char buf[SP_MAX_LINE];

    if (option == NULL || index < 0 || strnone(option->label)) {
        return NULL;
    }

    p = (char *)option->label;
    for (i = 0;; i++) {
        if (i == index) {
            label = p;
            break;
        }
        if ((p = strchr(p, '|')) == NULL) {
            if (i == 0 && index == 1) {
                return strclone(option->label);
            } else {
                return NULL;
            }
        }
        p++;
    }

    if (label != NULL) {
        spStrCopy(buf, sizeof(buf), label);
        
        if ((p = strchr(buf, '|')) != NULL) {
            *p = NUL;
        }
        spDebug(50, "xspGetOptionLabel", "label = %s\n", buf);
        
        return strclone(buf);
    } else {
        return NULL;
    }
}

void spSetOptionDescOffset(int offset)
{
    sp_option_desc_offset = offset;
    return;
}

void spSetHelpMessage(spBool *flag, const char *format, ...)
{
    va_list args;
    
    sp_help_flag = flag;
    va_start(args, format);
    spvsnprintf(sp_help_message, sizeof(sp_help_message), format, args);
    va_end(args);

    return;
}

const char *spGetHelpMessage(void)
{
    return sp_help_message;
}

static void printHelpHeader(spOptions options, const char *message)
{
    char buf[SP_MAX_LINE];
    
    if (options != NULL) {
        if (options->progname != NULL) {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(buf, sizeof(buf), "%s (%d)", options->progname, options->section);
#elif defined(HAVE_SNPRINTF)
            snprintf(buf, sizeof(buf), "%s (%d)", options->progname, options->section);
#else
            sprintf(buf, "%s (%d)", options->progname, options->section);
#endif
        }
        if (spgetstdout() == NULL) {
            printf("%-24s- %s\n", buf, message);
        } else {
            fprintf(spgetstdout(), "%-24s- %s\n", buf, message);
        }
    } else {
        if (spgetstdout() == NULL) {
            printf("%s\n", message);
        } else {
            fprintf(spgetstdout(), "%s\n", message);
        }
    }

    return;
}

void spPrintHelpHeader(const char *format, ...)
{
    va_list args;
    char message[SP_MAX_LINE];

    va_start(args, format);
    spvsnprintf(message, sizeof(message), format, args);
    va_end(args);

    printHelpHeader(sp_options, message);
    
    return;
}

/*
 *        print help
 */
void spPrintHelp(const char *format, ...)
{
    va_list args;
    char message[SP_MAX_LINE];

    va_start(args, format);
    spvsnprintf(message, sizeof(message), format, args);
    va_end(args);

    printHelpHeader(sp_options, message);
    spUsage();
}

/*
 *        print error
 */
void spPrintError(const char *format, ...)
{
    va_list args;
    char message[SP_MAX_LINE];

    va_start(args, format);
    spvsnprintf(message, sizeof(message), format, args);
    va_end(args);

    if (sp_options != NULL) {
        if (sp_options->progname != NULL) {
            spMessage("%s: %s\n", sp_options->progname, message);
        }
        spUsage();
    } else {
        spMessage("%s\n", message);
        spExit(1);
    }
}

#if 1
static int escapeStringForMan(char *obuf, int bufsize, const char *str)
{
    int i;
    int strsize;
    const char *ptr;

    ptr = str;
    strsize = bufsize - 1;

    for (i = 0; i < strsize && *ptr != NUL; i++) {
        if (*ptr == '-') {
            obuf[i] = '\\';
            ++i;
            if (i >= strsize) {
                --i;
            }
        }
        obuf[i] = *ptr;
        ++ptr;
    }
    obuf[i] = NUL;
    
    return i;
}

static spBool printOptionForMan(spOption *option)
{
    int len;
    char *p;
    char buf[SP_MAX_LINE];
    char flag[SP_MAX_LINE];
    char label[SP_MAX_LINE];
    
    if (option == NULL || strnone(option->flag) || strnone(option->desc))
        return SP_FALSE;
    
    if ((p = xspGetOptionLabel(option, 0)) != NULL) {
        spStrCopy(label, sizeof(label), p);
        xfree(p);
    } else {
#if defined(HAVE_SPRINTF_S)
        sprintf_s(label, sizeof(label), "<%s>", SP_USAGE_LABEL_STRING);
#elif defined(HAVE_SNPRINTF)
        snprintf(label, sizeof(label), "<%s>", SP_USAGE_LABEL_STRING);
#else
        sprintf(label, "<%s>", SP_USAGE_LABEL_STRING);
#endif
    }

    printf("\\fB");
    escapeStringForMan(flag, sizeof(flag), option->flag);
    printf("%s\\fR", flag);
    
    if (!strnone(option->subflag)) {
        printf(", \\fB");
        escapeStringForMan(flag, sizeof(flag), option->subflag);
        printf("%s\\fR", flag);
    }

    if (spEqType(option->type, SP_TYPE_BOOL)) {
        if (!strnone(option->label)) {
            if (option->value != NULL) {
                if (*(spBool *)option->value == SP_FALSE) {
                    printf(" (\\fI%s\\fR[False])", label);
                } else {
                    printf(" (\\fI%s\\fR[True])", label);
                }
            } else {
                printf(" (\\fI%s\\fR)", label);
            }
        }
    } else if (option->value != NULL && /*!strnone(option->def_value)*/option->def_value != NULL) {
        char value[SP_MAX_LINE];

        if (spGetOptionValueString(option, value, sizeof(value)) == SP_TRUE
            && !strnone(value)) {
            printf(" \\fI%s\\fR[%s]", label, value);
        } else {
            printf(" \\fI%s\\fR", label);
        }
    } else {
        if (label[0] == '[' || label[0] == '<' || label[0] == '{') {
            len = strlen(label);
            buf[0] = label[0]; buf[1] = NUL;
            printf(" %s\\fI", buf);
            spStrCopy(buf, sizeof(buf), "\\fR");
            buf[3] = label[len - 1]; buf[4] = NUL;
            label[len - 1] = NUL;
            printf("%s%s", &label[1], buf);
        } else {
            printf(" \\fI%s\\fR", label);
        }
    }
    printf("\n%s\n", option->desc);

    return SP_TRUE;
}

/*
 *        print man
 */
void spPrintMan(spOptions options, const char *description, const char *author)
{
    if (options != NULL) {
        int i;
        int len;
        char progname[SP_MAX_LINE];
        char upper_progname[SP_MAX_LINE];
        char buf[SP_MAX_LINE];
        char filename[SP_MAX_LINE];
        char files[SP_MAX_PATHNAME];

        if (options->progname != NULL) {
            spStrCopy(progname, sizeof(progname), options->progname);
        } else {
            spStrCopy(progname, sizeof(progname), "a.out");
        }

        if (spEqSuffix(progname, ".exe")) {
            len = strlen(progname);
            progname[len - 4] = NUL;
        }

        for (i = 0; progname[i] != NUL; i++) {
            if (islower(progname[i])) {
                upper_progname[i] = toupper(progname[i]);
            } else {
                upper_progname[i] = progname[i];
            }
        }
        upper_progname[i] = NUL;

        printf(".TH %s %d\n", upper_progname, options->section);
        printf(".SH NAME\n");
        printf("%s \\- ", progname);
        if (!strnone(sp_help_message)) {
            printf("%s\n", sp_help_message);
        } else {
            printf("execute %s\n", upper_progname);
        }

        spStrCopy(files, sizeof(files), "");
        spStrCopy(filename, sizeof(filename), "");

        for (i = 0; i < options->num_file; i++) {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(filename, sizeof(filename), "%s", options->labels[i]);
#elif defined(HAVE_SNPRINTF)
            snprintf(filename, sizeof(filename), "%s", options->labels[i]);
#else
            sprintf(filename, "%s", options->labels[i]);
#endif

            if (filename[0] == '[' || filename[0] == '<' || filename[0] == '{') {
                len = strlen(filename);
                buf[0] = ' '; buf[1] = filename[0]; buf[2] = NUL;
                spStrCat(files, sizeof(files), buf);
                spStrCat(files, sizeof(files), "\\fI");
                spStrCopy(buf, sizeof(buf), "\\fR");
                buf[3] = filename[len - 1]; buf[4] = NUL;
                filename[len - 1] = NUL;
                spStrCat(files, sizeof(files), &filename[1]);
                spStrCat(files, sizeof(files), buf);
            } else {
                spStrCat(files, sizeof(files), " <\\fI");
                spStrCat(files, sizeof(files), filename);
                spStrCat(files, sizeof(files), "\\fR>");
            }
        }
        
        printf(".SH SYNOPSIS\n");
        printf(".B %s\n", progname);
        if (options->num_option > 0) {
            printf("[\\fIoptions\\fR...]");
        }
        printf("%s\n", files);

        if (!strnone(description)) {
            printf(".SH DESCRIPTION\n");
#if 0
            printf(description);
            printf("\n");
#else
            printf("%s\n", description);
#endif
        }

        if (options->num_option > 0) {
            printf(".SH OPTIONS\n");
            for (i = 0; i < options->num_option; i++) {
                printf(".TP\n");
                printOptionForMan(&options->options[i]);
            }
        }
        
        if (!strnone(author)) {
            printf(".SH AUTHOR\n");
#if 0
            printf(author);
            printf("\n");
#else
            printf("%s\n", author);
#endif
        }
    }

    spExit(0);
    
    return;
}
#endif

spBool spIsGlobalSetup(const char *filename) {
    if (sp_setup_file != NULL && streq(spGetBaseName(sp_setup_file), spGetBaseName(filename))) {
        return SP_TRUE;
    }

    return SP_FALSE;
}

#if defined(MACOSX) || defined(IPHONE)
static spBool spReadOptionValueFromPreferences(spOption *option)
{
    spBool flag = SP_FALSE;
    char *label;
    CFStringRef cfskey;
    CFPropertyListRef plist;

    if ((label = xspGetOptionLabel(option, 0)) != NULL) {
        spDebug(80, "spReadOptionValueFromPreferences", "label = %s\n", label);
        if ((cfskey = CFStringCreateWithCString(kCFAllocatorDefault, label, kCFStringEncodingUTF8)) != NULL) {
            if ((plist = CFPreferencesCopyAppValue(cfskey, kCFPreferencesCurrentApplication)) != NULL) {
                flag = SP_TRUE;
                if (spEqType(option->type, SP_TYPE_LONG) && sizeof(long) == 8) {
                    SInt64 value;
                    CFNumberGetValue((CFNumberRef)plist, kCFNumberSInt64Type, &value);
                    spDebug(80, "spReadOptionValueFromPreferences", "SInt64 value = %ld\n", (long)value);
                    *(long *)option->value = value;
                    spDebug(80, "spReadOptionValueFromPreferences", "SInt64 option->value = %ld\n", (long)*(long *)option->value);
                } else if (spEqType(option->type, SP_TYPE_INT) || spEqType(option->type, SP_TYPE_LONG)) {
                    SInt32 value;
                    CFNumberGetValue((CFNumberRef)plist, kCFNumberSInt32Type, &value);
                    spDebug(80, "spReadOptionValueFromPreferences", "SInt32 value = %d\n", value);
                    if (spEqType(option->type, SP_TYPE_INT)) {
                        *(int *)option->value = value;
                    } else {
                        *(long *)option->value = value;
                    }
                } else if (spEqType(option->type, SP_TYPE_SHORT)) {
                    SInt16 value;
                    CFNumberGetValue((CFNumberRef)plist, kCFNumberSInt16Type, &value);
                    *(short *)option->value = value;
                } else if (spEqType(option->type, SP_TYPE_FLOAT)) {
                    Float32 value;
                    CFNumberGetValue((CFNumberRef)plist, kCFNumberFloat32Type, &value);
                    *(float *)option->value = value;
                } else if (spEqType(option->type, SP_TYPE_DOUBLE)) {
                    Float64 value;
                    CFNumberGetValue((CFNumberRef)plist, kCFNumberFloat64Type, &value);
                    *(double *)option->value = value;
                } else if (spEqType(option->type, SP_TYPE_BOOL)) {
                    *(spBool *)option->value = (CFBooleanGetValue((CFBooleanRef)plist) ? SP_TRUE : SP_FALSE);
                } else {
                    char *cstr;
                    
                    if ((cstr = (char *)CFStringGetCStringPtr((CFStringRef)plist, kCFStringEncodingUTF8)) != NULL) {
                        if (spEqType(option->type, SP_TYPE_CHAR)) {
                            *(char *)option->value = cstr[0];
                        } else if (spEqType(option->type, SP_TYPE_STRING)) {
                            if (*(char **)option->value != NULL) {
                                xfree(*(char **)option->value);
                            }
                            *(char **)option->value = strclone(cstr);
                        } else if (spEqType(option->type, SP_TYPE_STRING_A)) {
                            spStrCopy((char *)option->value, strlen(cstr) + 1, cstr);
                        } else {
                            flag = SP_FALSE;
                        }
                    }
                }
                
                CFRelease(plist);
            }
            CFRelease(cfskey);
        }
        xfree(label);
    }

    return flag;
}

static spBool spReadOptionsFromPreferences(spOptions options)
{
    int i;
    spBool flag = SP_FALSE;
    
    for (i = 0; i < options->num_option; i++) {
        if (options->changes == NULL || options->changes[i] != SP_TRUE) {
            spDebug(80, "spReadOptionsFromPreferences", "i = %d / %d\n", i, options->num_option);
            if (spReadOptionValueFromPreferences(&options->options[i])) {
                flag = SP_TRUE;
            }
        } else {
            spDebug(80, "spReadOptionsFromPreferences", "%d: already updated\n", i);
        }
    }

    return flag;
}

spBool spWriteOptionValueToPreferences(spOption *option)
{
    spBool flag = SP_FALSE;
    char *label;
    CFStringRef cfskey;
    CFPropertyListRef plist;

    if ((label = xspGetOptionLabel(option, 0)) != NULL) {
        if ((cfskey = CFStringCreateWithCString(kCFAllocatorDefault, label, kCFStringEncodingUTF8)) != NULL) {
            plist = NULL;
            
            if (spEqType(option->type, SP_TYPE_LONG) && sizeof(long) == 8) {
                SInt64 value;
                value = *(long *)option->value;
                plist = (CFPropertyListRef)CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt64Type, &value);
            } else if (spEqType(option->type, SP_TYPE_INT) || spEqType(option->type, SP_TYPE_LONG)) {
                SInt32 value;
                if (spEqType(option->type, SP_TYPE_INT)) {
                    value = *(int *)option->value;
                } else {
                    value = *(long *)option->value;
                }
                plist = (CFPropertyListRef)CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &value);
            } else if (spEqType(option->type, SP_TYPE_SHORT)) {
                SInt16 value;
                value = *(short *)option->value;
                plist = (CFPropertyListRef)CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt16Type, &value);
            } else if (spEqType(option->type, SP_TYPE_FLOAT)) {
                Float32 value;
                value = *(float *)option->value;
                plist = (CFPropertyListRef)CFNumberCreate(kCFAllocatorDefault, kCFNumberFloat32Type, &value);
            } else if (spEqType(option->type, SP_TYPE_DOUBLE)) {
                Float64 value;
                value = *(double *)option->value;
                plist = (CFPropertyListRef)CFNumberCreate(kCFAllocatorDefault, kCFNumberFloat64Type, &value);
            } else if (spEqType(option->type, SP_TYPE_BOOL)) {
                if (*(spBool *)option->value) {
                    plist = (CFPropertyListRef)kCFBooleanTrue;
                } else {
                    plist = (CFPropertyListRef)kCFBooleanFalse;
                }
            } else {
                char cstrbuf[2];
                char *cstr = NULL;
                    
                if (spEqType(option->type, SP_TYPE_CHAR)) {
                    cstrbuf[0] = *(char *)option->value;
                    cstrbuf[1] = NUL;
                    cstr = cstrbuf;
                } else if (spEqType(option->type, SP_TYPE_STRING)) {
                    cstr = *(char **)option->value;
                } else if (spEqType(option->type, SP_TYPE_STRING_A)) {
                    cstr = (char *)option->value;
                }

                if (cstr != NULL) {
                    plist = (CFPropertyListRef)CFStringCreateWithCString(kCFAllocatorDefault, cstr, kCFStringEncodingUTF8);
                }
            }

            if (plist != NULL) {
                flag = SP_TRUE;
                CFPreferencesSetAppValue(cfskey, plist, kCFPreferencesCurrentApplication);
                CFRelease(plist);
            }
            CFRelease(cfskey);
        }
        xfree(label);
    }

    return flag;
}

static spBool spWriteOptionsToPreferences(spOptions options)
{
    int i;
    spBool flag = SP_TRUE;
    
    for (i = 0; i < options->num_option; i++) {
        if (spWriteOptionValueToPreferences(&options->options[i]) == SP_FALSE) {
            flag = SP_FALSE;
        }
    }
    CFPreferencesAppSynchronize(kCFPreferencesCurrentApplication);

    return flag;
}
#endif

/*
 *        get option number of being equal to label
 */
static int labelEq(spOptions options, const char *label)
{
    int i;
    char *string;

    for (i = 0;; i++) {
        if (i >= options->num_option) {
            i = SP_UNKNOWN;
            break;
        }
        if ((string = xspGetOptionLabel(&options->options[i], 0)) != NULL) {
            if (streq(string, label)) {
                xfree(string);
                break;
            }
            xfree(string);
        }
    }

    return i;
}

/*
 *        read setup file
 */
spBool spReadSetup(const char *filename, spOptions options)
{
    int j;
    char *exactname;
    char name[SP_MAX_SETUP_NAME];
    char value[SP_MAX_SETUP_VALUE];
    char line[SP_MAX_SETUP_LINE];
    FILE *fp;
    spBool flag;

    if (strnone(filename) || options == NULL)
        return SP_FALSE;

    spDebug(80, "spReadSetup", "filename = %s\n", filename);
    
    if (spIsGlobalSetup(filename)) {
#if defined(MACOSX) || defined(IPHONE)
        return spReadOptionsFromPreferences(options);
#else
        filename = sp_setup_file;
#endif
    }

    spDebug(80, "spReadSetup", "filename = %s, sp_setup_file = %s\n", filename, sp_setup_file);
    
    exactname = xspGetExactName(filename);

    if (NULL == (fp = spOpenFile(exactname, "r"))) {
        spDebug(80, "spReadSetup", "can't open %s\n", exactname);
        flag = SP_FALSE;
    } else {
        flag = SP_TRUE;
        while (fgetnline(line, SP_MAX_SETUP_LINE, fp) != EOF) {
            sscansetup(line, name, value);
            spDebug(100, "spReadSetup", "name = %s, value = %s\n", name, value);
            if (!strnone(value) && (j = labelEq(options, name)) >= 0) {
                spDebug(100, "spReadSetup", "j = %d\n", j);
                if (options->changes == NULL || options->changes[j] != SP_TRUE) {
                    spDebug(100, "spReadSetup", "call spConvertOptionValue\n");
                    spConvertOptionValue(&options->options[j], value);
                } else {
                    spDebug(80, "spReadSetup", "%s: already updated\n", name);
                }
            }
            name[0] = NUL;
            value[0] = NUL;
        }
        spCloseFile(fp);
    }

    xfree(exactname);
    
    return flag;
}

/*
 *        write setup file
 */
spBool spWriteSetup(const char *filename, spOptions options)
{
    int i;
    char *label;
    char *string;
    char *exactname;
    FILE *fp;

    if (strnone(filename) || options == NULL)
        return SP_FALSE;

    if (spIsGlobalSetup(filename)) {
#if defined(MACOSX) || defined(IPHONE)
        return spWriteOptionsToPreferences(options);
#else
        filename = sp_setup_file;
#endif
    }

    exactname = xspGetExactName(filename);
    
    spDebug(40, "spWriteSetup", "exactname = %s\n", exactname);

    if (NULL == (fp = spOpenFile(exactname, "w"))) {
        spMessage("Can't open setup file: %s\n", exactname);
        return SP_FALSE;
    }
    
    for (i = 0; i < options->num_option; i++) {
        if ((label = xspGetOptionLabel(&options->options[i], 0)) == NULL)
            continue;
        
        fprintf(fp, "%s ", label);
        xfree(label);
        
        if (spEqType(options->options[i].type, SP_TYPE_BOOL)) {
            fprintf(fp, "%s", spBoolToStr(*(spBool *)options->options[i].value));
        } else if (spEqType(options->options[i].type, SP_TYPE_INT)) {
            fprintf(fp, "%d", *(int *)options->options[i].value);
        } else if (spEqType(options->options[i].type, SP_TYPE_SHORT)) {
            fprintf(fp, "%d", *(short *)options->options[i].value);
        } else if (spEqType(options->options[i].type, SP_TYPE_LONG)) {
            fprintf(fp, "%ld", *(long *)options->options[i].value);
        } else if (spEqType(options->options[i].type, SP_TYPE_FLOAT)) {
            fprintf(fp, "%f", *(float *)options->options[i].value);
        } else if (spEqType(options->options[i].type, SP_TYPE_DOUBLE)) {
            fprintf(fp, "%f", *(double *)options->options[i].value);
        } else if (spEqType(options->options[i].type, SP_TYPE_CHAR)) {
            fprintf(fp, "%c", *(char *)options->options[i].value);
        } else if (spEqType(options->options[i].type,  SP_TYPE_STRING)
                   || spEqType(options->options[i].type,  SP_TYPE_STRING_A)) {
            if (spEqType(options->options[i].type,  SP_TYPE_STRING)) {
                string = *(char **)options->options[i].value;
            } else {
                string = (char *)options->options[i].value;
            }
            if (strnone(string)) {
                fprintf(fp, "  ");
            } else {
                fputline(string, fp);
            }
        }
        fprintf(fp, "\n");
    }
    spCloseFile(fp);

    xfree(exactname);

    spDebug(40, "spWriteSetup", "done\n");
    
    return SP_TRUE;
}

spBool spGetOptionValueString(spOption *option, char *buf, int buf_size)
{
    if (option->value == NULL) return SP_FALSE;
    
    if (spEqType(option->type, SP_TYPE_BOOL)) {
        if (*(spBool *)option->value == SP_FALSE) {
            spStrCopy(buf, buf_size, "False");
        } else {
            spStrCopy(buf, buf_size, "True");
        }
    } else {
        if (spEqType(option->type, SP_TYPE_INT)) {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(buf, buf_size, "%d", *(int *)option->value);
#elif defined(HAVE_SNPRINTF)
            snprintf(buf, buf_size, "%d", *(int *)option->value);
#else
            sprintf(buf, "%d", *(int *)option->value);
#endif
        } else if (spEqType(option->type, SP_TYPE_SHORT)) {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(buf, buf_size, "%d", *(short *)option->value);
#elif defined(HAVE_SNPRINTF)
            snprintf(buf, buf_size, "%d", *(short *)option->value);
#else
            sprintf(buf, "%d", *(short *)option->value);
#endif
        } else if (spEqType(option->type, SP_TYPE_LONG)) {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(buf, buf_size, "%ld", *(long *)option->value);
#elif defined(HAVE_SNPRINTF)
            snprintf(buf, buf_size, "%ld", *(long *)option->value);
#else
            sprintf(buf, "%ld", *(long *)option->value);
#endif
        } else if (spEqType(option->type, SP_TYPE_FLOAT) ||
                   spEqType(option->type, SP_TYPE_DOUBLE)) {
            int j;
            char value[SP_MAX_LINE];
            
            if (spEqType(option->type, SP_TYPE_FLOAT)) {
#if defined(HAVE_SPRINTF_S)
                sprintf_s(value, sizeof(value), "%f", *(float *)option->value);
#elif defined(HAVE_SNPRINTF)
                snprintf(value, sizeof(value), "%f", *(float *)option->value);
#else
                sprintf(value, "%f", *(float *)option->value);
#endif
            } else {
#if defined(HAVE_SPRINTF_S)
                sprintf_s(value, sizeof(value), "%f", *(double *)option->value);
#elif defined(HAVE_SNPRINTF)
                snprintf(value, sizeof(value), "%f", *(double *)option->value);
#else
                sprintf(value, "%f", *(double *)option->value);
#endif
            }
            for (j = (int)strlen(value) - 1; j >= 0; j--) {
                if (value[j] == '.') {
                    value[MIN(j + 2, (int)strlen(value))] = NUL;
                    break;
                } else if (value[j] != '0') {
                    value[j + 1] = NUL;
                    break;
                }
            }
            spStrCopy(buf, buf_size, value);
        } else if (spEqType(option->type, SP_TYPE_CHAR)) {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(buf, buf_size, "%c", *(char *)option->value);
#elif defined(HAVE_SNPRINTF)
            snprintf(buf, buf_size, "%c", *(char *)option->value);
#else
            sprintf(buf, "%c", *(char *)option->value);
#endif
        } else if (spEqType(option->type, SP_TYPE_STRING)) {
            if (strnone(*(char **)option->value)) {
                buf[0] = NUL;
            } else {
                spStrCopy(buf, buf_size, *(char **)option->value);
            }
        } else if (spEqType(option->type, SP_TYPE_STRING_A)) {
            if (strnone((char *)option->value)) {
                buf[0] = NUL;
            } else {
                spStrCopy(buf, buf_size, (char *)option->value);
            }
        } else {
            spWarning("Unknown option data type.\n");
            return SP_FALSE;
        }
    }

    return SP_TRUE;
}

static spBool printOption(spOption *option, int desc_offset)
{
    char *p;
    char format[SP_MAX_MESSAGE];
    char buf[SP_MAX_PATHNAME];
    char flag[SP_MAX_LINE];
    char label[SP_MAX_LINE];
    
    if (option == NULL || strnone(option->flag) || strnone(option->desc))
        return SP_FALSE;
    
    if ((p = xspGetOptionLabel(option, 0)) != NULL) {
        spStrCopy(label, sizeof(label), p);
        xfree(p);
    } else {
        spStrCopy(label, sizeof(label), SP_USAGE_LABEL_STRING);
    }
    
    if (!strnone(option->subflag)) {
#if defined(HAVE_SPRINTF_S)
        sprintf_s(flag, sizeof(flag), "%s %s", option->flag, option->subflag);
#elif defined(HAVE_SNPRINTF)
        snprintf(flag, sizeof(flag), "%s %s", option->flag, option->subflag);
#else
        sprintf(flag, "%s %s", option->flag, option->subflag);
#endif
    } else {
#if defined(HAVE_SPRINTF_S)
        sprintf_s(flag, sizeof(flag), "%s", option->flag);
#elif defined(HAVE_SNPRINTF)
        snprintf(flag, sizeof(flag), "%s", option->flag);
#else
        sprintf(flag, "%s", option->flag);
#endif
    }

#if defined(HAVE_SPRINTF_S)
    sprintf_s(format, sizeof(format), "    %%-%ds: %%s\n", desc_offset);
#elif defined(HAVE_SNPRINTF)
    snprintf(format, sizeof(format), "    %%-%ds: %%s\n", desc_offset);
#else
    sprintf(format, "    %%-%ds: %%s\n", desc_offset);
#endif
    
    if (spEqType(option->type, SP_TYPE_BOOL)) {
        if (!strnone(option->label)) {
            if (option->value != NULL) {
                if (*(spBool *)option->value == SP_FALSE) {
#if defined(HAVE_SPRINTF_S)
                    sprintf_s(buf, sizeof(buf), "%s  (%s[False])", flag, label);
#elif defined(HAVE_SNPRINTF)
                    snprintf(buf, sizeof(buf), "%s  (%s[False])", flag, label);
#else
                    sprintf(buf, "%s  (%s[False])", flag, label);
#endif
                } else {
#if defined(HAVE_SPRINTF_S)
                    sprintf_s(buf, sizeof(buf), "%s  (%s[True])", flag, label);
#elif defined(HAVE_SNPRINTF)
                    snprintf(buf, sizeof(buf), "%s  (%s[True])", flag, label);
#else
                    sprintf(buf, "%s  (%s[True])", flag, label);
#endif
                }
            } else {
#if defined(HAVE_SPRINTF_S)
                sprintf_s(buf, sizeof(buf), "%s  (%s)", flag, label);
#elif defined(HAVE_SNPRINTF)
                snprintf(buf, sizeof(buf), "%s  (%s)", flag, label);
#else
                sprintf(buf, "%s  (%s)", flag, label);
#endif
            }
            if (spgetstdout() == NULL) {
                printf(format, buf, option->desc);
            } else {
                fprintf(spgetstdout(), format, buf, option->desc);
            }
        } else {
            if (spgetstdout() == NULL) {
                printf(format, flag, option->desc);
            } else {
                fprintf(spgetstdout(), format, flag, option->desc);
            }
        }
    } else if (option->value != NULL && /*!strnone(option->def_value)*/option->def_value != NULL) {
        char value[SP_MAX_LINE];

        if (spGetOptionValueString(option, value, sizeof(value)) == SP_TRUE
            && !strnone(value)) {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(buf, sizeof(buf), "%s %s[%s]", flag, label, value);
#elif defined(HAVE_SNPRINTF)
            snprintf(buf, sizeof(buf), "%s %s[%s]", flag, label, value);
#else
            sprintf(buf, "%s %s[%s]", flag, label, value);
#endif
        } else {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(buf, sizeof(buf), "%s %s", flag, label);
#elif defined(HAVE_SNPRINTF)
            snprintf(buf, sizeof(buf), "%s %s", flag, label);
#else
            sprintf(buf, "%s %s", flag, label);
#endif
        }
        
        if (spgetstdout() == NULL) {
            printf(format, buf, option->desc);
        } else {
            fprintf(spgetstdout(), format, buf, option->desc);
        }
    } else {
#if defined(HAVE_SPRINTF_S)
        sprintf_s(buf, sizeof(buf), "%s %s", flag, label);
#elif defined(HAVE_SNPRINTF)
        snprintf(buf, sizeof(buf), "%s %s", flag, label);
#else
        sprintf(buf, "%s %s", flag, label);
#endif
        if (spgetstdout() == NULL) {
            printf(format, buf, option->desc);
        } else {
            fprintf(spgetstdout(), format, buf, option->desc);
        }
    }

    return SP_TRUE;
}

spBool spPrintOption(spOption *option)
{
    return printOption(option, sp_option_desc_offset);
}

spBool spPrintOptions(spOptions options)
{
    int i;
    
    if (options == NULL) return SP_FALSE;
    
    for (i = 0; i < options->num_option; i++) {
        spPrintOption(&options->options[i]);
    }

    return SP_TRUE;
}

void spPrintUsageHeader(void)
{
    int i;
    spOptions options = sp_options;
    char buf[SP_MAX_PATHNAME];
    char filename[SP_MAX_LINE];

    if (options == NULL) return;
    
    spStrCopy(buf, sizeof(buf), "");
    spStrCopy(filename, sizeof(filename), "");
    
    spDebug(40, "spPrintUsageHeader", "num_file = %d, num_option = %d\n",
            options->num_file, options->num_option);
    
    for (i = 0; i < options->num_file; i++) {
#if defined(HAVE_SPRINTF_S)
        sprintf_s(filename, sizeof(filename), " %s", options->labels[i]);
#elif defined(HAVE_SNPRINTF)
        snprintf(filename, sizeof(filename), " %s", options->labels[i]);
#else
        sprintf(filename, " %s", options->labels[i]);
#endif
        spStrCat(buf, sizeof(buf), filename);
    }

    if (options->num_option <= 0) {
        if (options->progname != NULL) {
            if (spgetstdout() == NULL) {
                printf("usage: %s%s\n", options->progname, buf);
            } else {
                fprintf(spgetstdout(), "usage: %s%s\n", options->progname, buf);
            }
        }
    } else {
        if (options->progname != NULL) {
            if (spgetstdout() == NULL) {
                printf("usage: %s [options...]%s\n", options->progname, buf);
                printf("options:\n");
            } else {
                fprintf(spgetstdout(), "usage: %s [options...]%s\n", options->progname, buf);
                fprintf(spgetstdout(), "options:\n");
            }
        }
    }

    return;
}

void spPrintUsage(void)
{
    if (sp_options != NULL) {
        spPrintOptions(sp_options);
        if (spgetstdout() == NULL) {
            printf("\n");
        } else {
            fprintf(spgetstdout(), "\n");
        }
    }
    spExit(1);

    return;
}

void spUsage(void)
{
    spPrintUsageHeader();
    spPrintUsage();
    
    return;
}

#if defined(_WIN32)
static void *sp_main_hinstance = NULL;

void spSetHInstanceWin(void *hinstance, unsigned long reason)
{
    sp_main_hinstance = hinstance;
}

void *spGetHInstanceWin(void)
{
    return sp_main_hinstance;
}
#endif
